#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pickle

import numpy

try:
    import numba
    import llvmlite.llvmpy.core
except ImportError:
    numba = None

import oamap.generator
import oamap.proxy

if numba is not None:
    ################################################################ Cache

    class CacheType(numba.types.Type):
        def __init__(self):
            super(CacheType, self).__init__(name="CacheType")

    cachetype = CacheType()

    @numba.extending.register_model(CacheType)
    class CacheModel(numba.datamodel.models.StructModel):
        def __init__(self, dmm, fe_type):
            members = [("cache", numba.types.intp),
                       ("ptr", numba.types.intp),
                       ("len", numba.types.intp)]
            super(CacheModel, self).__init__(dmm, fe_type, members)

    def unbox_cache(cache_obj, c):
        entercompiled_fcn = c.pyapi.object_getattr_string(cache_obj, "entercompiled")
        pair_obj = c.pyapi.call_function_objargs(entercompiled_fcn, ())
        ptr_obj = c.pyapi.tuple_getitem(pair_obj, 0)
        len_obj = c.pyapi.tuple_getitem(pair_obj, 1)
        ptr_val = c.pyapi.number_as_ssize_t(ptr_obj)
        len_val = c.pyapi.number_as_ssize_t(len_obj)

        cache = numba.cgutils.create_struct_proxy(cachetype)(c.context, c.builder)
        cache.cache = c.builder.ptrtoint(cache_obj, llvmlite.llvmpy.core.Type.int(numba.types.intp.bitwidth))
        cache.ptr = c.builder.ptrtoint(ptr_val, llvmlite.llvmpy.core.Type.int(numba.types.intp.bitwidth))
        cache.len = c.builder.ptrtoint(len_val, llvmlite.llvmpy.core.Type.int(numba.types.intp.bitwidth))

        c.pyapi.decref(cache_obj)       # not this one: this one is decrefed by the calling function
        c.pyapi.decref(pair_obj)
        c.pyapi.decref(ptr_obj)
        c.pyapi.decref(len_obj)
        
        return cache._getvalue()

    def box_cache(context, builder, pyapi, cache_val):
        cache = numba.cgutils.create_struct_proxy(cachetype)(context, builder, value=cache_val)
        return builder.inttoptr(cache.cache, pyapi.pyobj)

    ################################################################ general routines for all types

    def typeof_generator(generator):
        if isinstance(generator, oamap.generator.MaskedPrimitiveGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.PrimitiveGenerator):
            if generator.dims == ():
                return numba.from_dtype(generator.dtype)
            else:
                raise NotImplementedError

        elif isinstance(generator, oamap.generator.MaskedListGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.ListGenerator):
            return ListProxyNumbaType(generator)

        elif isinstance(generator, oamap.generator.MaskedUnionGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.UnionGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.MaskedRecordGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.RecordGenerator):
            return RecordProxyNumbaType(generator)

        elif isinstance(generator, oamap.generator.MaskedTupleGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.TupleGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.MaskedPointerGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.PointerGenerator):
            raise NotImplementedError

        else:
            raise AssertionError("unrecognized generator type: {0} ({1})".format(generator.__class__, repr(generator)))

    @numba.extending.typeof_impl.register(oamap.proxy.Proxy)
    def typeof_proxy(val, c):
        return typeof_generator(val._generator)

    def getarray(arrays, name, cache, cacheidx, dtype, dims):
        try:
            array = arrays[name]
            if not isinstance(array, numpy.ndarray):
                raise TypeError("arrays[{0}] returned a {1} ({2}) instead of a Numpy array".format(repr(name), type(array), repr(array)))
            if dtype is not None and array.dtype != dtype:
                raise TypeError("arrays[{0}] returned an array of type {1} instead of {2}".format(repr(name), array.dtype, dtype))
            if dims is not None and array.shape[1:] != dims:
                raise TypeError("arrays[{0}] returned an array with shape[1:] {1} instead of {2}".format(repr(name), array.shape[1:], dims))
            cache.arraylist[cacheidx] = array
            cache.ptr[cacheidx] = array.ctypes.data
            cache.len[cacheidx] = array.shape[0]
        except:
            return False
        else:
            return True

    def constint(value, bits=64):
        return llvmlite.llvmpy.core.Constant.int(llvmlite.llvmpy.core.Type.int(bits), value)

    def atidx(context, builder, ptr, idx):
        pos = builder.add(ptr, builder.mul(idx, constint(numba.types.intp.bitwidth // 8)))
        posptr = builder.inttoptr(pos, llvmlite.llvmpy.core.Type.pointer(context.get_value_type(numba.types.intp)))
        return numba.targets.arrayobj.load_item(context, builder, numba.types.intp[:], posptr)

    def runtimeerror(context, builder, pyapi, case, message):
        with builder.if_then(case, likely=False):
            exc = pyapi.serialize_object(RuntimeError(message))
            excptr = context.call_conv._get_excinfo_argument(builder.function)
            builder.store(exc, excptr)
            builder.ret(numba.targets.callconv.RETCODE_USEREXC)

    def ensure(context, builder, pyapi, ptr, arrays, name, cache, cacheidx, dtype, dims):
        with builder.if_then(builder.not_(context.is_true(builder, numba.types.int8, ptr)), likely=False):
            getarray_fcn = pyapi.unserialize(pyapi.serialize_object(getarray))
            arrays_obj = builder.inttoptr(arrays, pyapi.pyobj)
            name_obj = pyapi.unserialize(pyapi.serialize_object(name))
            cache_obj = box_cache(context, builder, pyapi, cache)
            cacheidx_obj = pyapi.long_from_long(cacheidx)
            dtype_obj = pyapi.unserialize(pyapi.serialize_object(dtype))
            dims_obj = pyapi.unserialize(pyapi.serialize_object(dims))
            ok_obj = pyapi.call_function_objargs(getarray_fcn, (arrays_obj, name_obj, cache_obj, cacheidx_obj, dtype_obj, dims_obj))
            ok = pyapi.object_istrue(ok_obj)
            pyapi.decref(cacheidx_obj)
            pyapi.decref(dtype_obj)
            pyapi.decref(dims_obj)
            pyapi.decref(ok_obj)
            runtimeerror(context, builder, pyapi, builder.not_(context.is_true(builder, numba.types.intc, ok)), "an exception occurred while trying to load an array")

    def arrayitem(context, builder, ptr, at, dtype):
        pos = builder.add(ptr, builder.mul(at, constint(dtype.itemsize)))
        posptr = builder.inttoptr(pos, llvmlite.llvmpy.core.Type.pointer(context.get_value_type(numba.from_dtype(dtype))))
        return numba.targets.arrayobj.load_item(context, builder, numba.from_dtype(dtype)[:], posptr)

    def castint(builder, val, bits=64):
        if val.type.width < bits:
            return builder.zext(val, llvmlite.llvmpy.core.Type.int(bits))
        elif val.type.width > bits:
            return builder.trunc(val, llvmlite.llvmpy.core.Type.int(bits))
        else:
            return builder.bitcast(val, llvmlite.llvmpy.core.Type.int(bits))

    def generate(context, builder, pyapi, arrays, cache, generator, at):
        cachestruct = numba.cgutils.create_struct_proxy(cachetype)(context, builder, value=cache)

        if isinstance(generator, oamap.generator.MaskedPrimitiveGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.PrimitiveGenerator):
            dataidx = constint(generator.dataidx)
            dataptr = atidx(context, builder, cachestruct.ptr, dataidx)
            ensure(context, builder, pyapi, dataptr, arrays, generator.data, cache, dataidx, generator.dtype, generator.dims)

            dataptr = atidx(context, builder, cachestruct.ptr, dataidx)
            datalen = atidx(context, builder, cachestruct.len, dataidx)
            runtimeerror(context, builder, pyapi, builder.icmp_unsigned(">=", at, datalen), "PrimitiveProxy data array index out of range")

            return arrayitem(context, builder, dataptr, at, generator.dtype)

        elif isinstance(generator, oamap.generator.MaskedListGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.ListGenerator):
            startsidx = constint(generator.startsidx)
            startsptr = atidx(context, builder, cachestruct.ptr, startsidx)
            ensure(context, builder, pyapi, startsptr, arrays, generator.starts, cache, startsidx, generator.dtype, ())

            stopsidx = constint(generator.stopsidx)
            stopsptr = atidx(context, builder, cachestruct.ptr, stopsidx)
            ensure(context, builder, pyapi, stopsptr, arrays, generator.stops, cache, stopsidx, generator.dtype, ())

            startsptr = atidx(context, builder, cachestruct.ptr, startsidx)
            startslen = atidx(context, builder, cachestruct.len, startsidx)
            runtimeerror(context, builder, pyapi, builder.icmp_unsigned(">=", at, startslen), "ListProxy starts array index out of range")

            stopsptr = atidx(context, builder, cachestruct.ptr, stopsidx)
            stopslen = atidx(context, builder, cachestruct.len, stopsidx)
            runtimeerror(context, builder, pyapi, builder.icmp_unsigned(">=", at, stopslen), "ListProxy stops array index out of range")

            listproxy = numba.cgutils.create_struct_proxy(typeof_generator(generator))(context, builder)
            listproxy.arrays = arrays
            listproxy.cache = cache
            listproxy.start = castint(builder, arrayitem(context, builder, startsptr, at, generator.dtype))
            listproxy.stop = castint(builder, arrayitem(context, builder, stopsptr,  at, generator.dtype))
            listproxy.step = constint(1)
            return listproxy._getvalue()

        elif isinstance(generator, oamap.generator.MaskedUnionGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.UnionGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.MaskedRecordGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.RecordGenerator):
            recordproxy = numba.cgutils.create_struct_proxy(typeof_generator(generator))(context, builder)
            recordproxy.arrays = arrays
            recordproxy.cache = cache
            recordproxy.index = at
            return recordproxy._getvalue()

        elif isinstance(generator, oamap.generator.MaskedTupleGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.TupleGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.MaskedPointerGenerator):
            raise NotImplementedError

        elif isinstance(generator, oamap.generator.PointerGenerator):
            raise NotImplementedError

        else:
            raise NotImplementedError

    ################################################################ ListProxy

    class ListProxyNumbaType(numba.types.Type):
        def __init__(self, generator):
            self.generator = generator
            super(ListProxyNumbaType, self).__init__(name=self.generator._uniquestr)

    @numba.extending.register_model(ListProxyNumbaType)
    class ListProxyModel(numba.datamodel.models.StructModel):
        def __init__(self, dmm, fe_type):
            members = [("arrays", numba.types.intp),
                       ("cache", cachetype),
                       ("start", numba.types.int64),
                       ("stop", numba.types.int64),
                       ("step", numba.types.int64)]
            super(ListProxyModel, self).__init__(dmm, fe_type, members)

    @numba.typing.templates.infer
    class ListProxyGetItem(numba.typing.templates.AbstractTemplate):
        key = "getitem"
        def generic(self, args, kwds):
            tpe, idx = args
            if isinstance(tpe, ListProxyNumbaType):
                if isinstance(idx, numba.types.Integer):
                    return typeof_generator(tpe.generator.content)(tpe, idx)

    @numba.extending.lower_builtin("getitem", ListProxyNumbaType, numba.types.Integer)
    def listproxy_getitem(context, builder, sig, args):
        listtpe, indextpe = sig.args
        listval, indexval = args

        pyapi = context.get_python_api(builder)
        listproxy = numba.cgutils.create_struct_proxy(listtpe)(context, builder, value=listval)

        lenself = builder.sub(listproxy.stop, listproxy.start)
        zero = constint(0)
        normptr = numba.cgutils.alloca_once(builder, llvmlite.llvmpy.core.Type.int(64))
        builder.store(indexval, normptr)
        with builder.if_then(builder.icmp_signed("<", indexval, zero)):
            builder.store(builder.add(indexval, lenself), normptr)

        normval = builder.load(normptr)
        with builder.if_then(builder.or_(builder.icmp_signed("<", normval, zero),
                                         builder.icmp_signed(">=", normval, lenself)), likely=False):
            context.call_conv.return_user_exc(builder, IndexError, ("index out of bounds",))

        at = builder.add(listproxy.start, builder.mul(listproxy.step, normval))
        return generate(context, builder, pyapi, listproxy.arrays, listproxy.cache, listtpe.generator.content, at)

    @numba.extending.unbox(ListProxyNumbaType)
    def unbox_listproxy(typ, obj, c):
        arrays_obj = c.pyapi.object_getattr_string(obj, "_arrays")
        cache_obj = c.pyapi.object_getattr_string(obj, "_cache")
        start_obj = c.pyapi.object_getattr_string(obj, "_start")
        stop_obj = c.pyapi.object_getattr_string(obj, "_stop")
        step_obj = c.pyapi.object_getattr_string(obj, "_step")

        listproxy = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder)
        listproxy.arrays = c.builder.ptrtoint(arrays_obj, llvmlite.llvmpy.core.Type.int(numba.types.intp.bitwidth))
        listproxy.cache = unbox_cache(cache_obj, c)
        listproxy.start = c.pyapi.long_as_longlong(start_obj)
        listproxy.stop = c.pyapi.long_as_longlong(stop_obj)
        listproxy.step = c.pyapi.long_as_longlong(step_obj)

        c.pyapi.decref(arrays_obj)
        c.pyapi.decref(cache_obj)
        c.pyapi.decref(start_obj)
        c.pyapi.decref(stop_obj)
        c.pyapi.decref(step_obj)

        is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
        return numba.extending.NativeValue(listproxy._getvalue(), is_error=is_error)

    @numba.extending.box(ListProxyNumbaType)
    def box_listproxy(typ, val, c):
        listproxy = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
        arrays_obj = c.builder.inttoptr(listproxy.arrays, c.pyapi.pyobj)
        cache_obj = box_cache(c.context, c.builder, c.pyapi, listproxy.cache)
        start_obj = c.pyapi.long_from_longlong(listproxy.start)
        stop_obj = c.pyapi.long_from_longlong(listproxy.stop)
        step_obj = c.pyapi.long_from_longlong(listproxy.step)

        listproxy_cls = c.pyapi.unserialize(c.pyapi.serialize_object(oamap.proxy.ListProxy))
        generator_obj = c.pyapi.unserialize(c.pyapi.serialize_object(typ.generator))

        out = c.pyapi.call_function_objargs(listproxy_cls, (generator_obj, arrays_obj, cache_obj, start_obj, stop_obj, step_obj))

        # c.pyapi.decref(arrays_obj)      # this reference is exported
        # c.pyapi.decref(cache_obj)       # this reference is exported
        c.pyapi.decref(start_obj)
        c.pyapi.decref(stop_obj)
        c.pyapi.decref(step_obj)
        c.pyapi.decref(listproxy_cls)
        # c.pyapi.decref(generator_obj)   # this reference is exported
        return out

    ################################################################ RecordProxy

    class RecordProxyNumbaType(numba.types.Type):
        def __init__(self, generator):
            self.generator = generator
            super(RecordProxyNumbaType, self).__init__(name=self.generator._uniquestr)

    @numba.extending.register_model(RecordProxyNumbaType)
    class RecordProxyModel(numba.datamodel.models.StructModel):
        def __init__(self, dmm, fe_type):
            members = [("arrays", numba.types.intp),
                       ("cache", cachetype),
                       ("index", numba.types.int64)]
            super(RecordProxyModel, self).__init__(dmm, fe_type, members)

    @numba.extending.infer_getattr
    class StructAttribute(numba.typing.templates.AttributeTemplate):
        key = RecordProxyNumbaType
        def generic_resolve(self, typ, attr):
            fieldgenerator = typ.generator.fields.get(attr, None)
            if fieldgenerator is not None:
                return typeof_generator(fieldgenerator)
            else:
                raise AttributeError("{0} object has no attribute {1}".format(repr("Record" if typ.generator.name is None else typ.generator.name), repr(attr)))

    @numba.extending.lower_getattr_generic(RecordProxyNumbaType)
    def recordproxy_getattr(context, builder, typ, val, attr):
        pyapi = context.get_python_api(builder)
        recordproxy = numba.cgutils.create_struct_proxy(typ)(context, builder, value=val)
        return generate(context, builder, pyapi, recordproxy.arrays, recordproxy.cache, typ.generator.fields[attr], recordproxy.index)

    @numba.extending.unbox(RecordProxyNumbaType)
    def unbox_recordproxy(typ, obj, c):
        arrays_obj = c.pyapi.object_getattr_string(obj, "_arrays")
        cache_obj = c.pyapi.object_getattr_string(obj, "_cache")
        index_obj = c.pyapi.object_getattr_string(obj, "_index")

        recordproxy = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder)
        recordproxy.arrays = c.builder.ptrtoint(arrays_obj, llvmlite.llvmpy.core.Type.int(numba.types.intp.bitwidth))
        recordproxy.cache = unbox_cache(cache_obj, c)
        recordproxy.index = c.pyapi.long_as_longlong(index_obj)

        c.pyapi.decref(arrays_obj)
        c.pyapi.decref(cache_obj)
        c.pyapi.decref(index_obj)

        is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
        return numba.extending.NativeValue(recordproxy._getvalue(), is_error=is_error)

    @numba.extending.box(RecordProxyNumbaType)
    def box_recordproxy(typ, val, c):
        recordproxy = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
        arrays_obj = c.builder.inttoptr(recordproxy.arrays, c.pyapi.pyobj)
        cache_obj = box_cache(c.context, c.builder, c.pyapi, recordproxy.cache)
        index_obj = c.pyapi.long_from_longlong(recordproxy.index)

        recordproxy_cls = c.pyapi.unserialize(c.pyapi.serialize_object(oamap.proxy.RecordProxy))
        generator_obj = c.pyapi.unserialize(c.pyapi.serialize_object(typ.generator))

        out = c.pyapi.call_function_objargs(recordproxy_cls, (generator_obj, arrays_obj, cache_obj, index_obj))

        # c.pyapi.decref(arrays_obj)      # this reference is exported
        # c.pyapi.decref(cache_obj)       # this reference is exported
        c.pyapi.decref(index_obj)
        c.pyapi.decref(recordproxy_cls)
        # c.pyapi.decref(generator_obj)   # this reference is exported
        return out



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys

import numpy

import oamap.proxy

if sys.version_info[0] > 2:
    basestring = str

# array cache, so that arrays are only loaded once (might be an expensive network operation)
class Cache(object):
    def __init__(self, cachelen):
        self.arraylist = [None] * cachelen
        self.ptr = numpy.zeros(cachelen, dtype=numpy.intp)   # these arrays are only set and used in compiled code
        self.len = numpy.zeros(cachelen, dtype=numpy.intp)

    def entercompiled(self):
        for i, x in enumerate(self.arraylist):
            if x is None:
                self.ptr[i] = 0
                self.len[i] = 0
            else:
                if not isinstance(x, numpy.ndarray):
                    raise TypeError("all arrays must have numpy.ndarray type for use in compiled code")
                self.ptr[i] = x.ctypes.data
                self.len[i] = x.shape[0]
        return self.ptr.ctypes.data, self.len.ctypes.data

# base class of all runtime-object generators (one for each type)
class Generator(object):
    def _toarray(self, maybearray, dtype):
        if isinstance(maybearray, numpy.ndarray):
            return maybearray
        else:
            return numpy.array(maybearray, dtype=dtype)

    def _getarray(self, arrays, name, cache, cacheidx, dtype, dims):
        if cache.arraylist[cacheidx] is None:
            try:
                array = self._toarray(arrays[name], dtype)
            except KeyError as err:
                array = self._fallback(arrays, cache, name, err)
            cache.arraylist[cacheidx] = array
            if cache.arraylist[cacheidx].dtype != dtype:
                array = numpy.array(array, dtype=dtype)
            if cache.arraylist[cacheidx].shape[1:] != dims:
                raise TypeError("arrays[{0}].shape[1:] is {1} but expected {2}".format(repr(name), cache.arraylist[cacheidx].shape[1:], dims))
        return cache.arraylist[cacheidx]

    def _fallback(self, arrays, cache, name, err):
        raise err

    def __init__(self, name, derivedname, schema):
        self.name = name
        self.derivedname = derivedname
        self.schema = schema

    def __call__(self, arrays):
        return self._generate(arrays, 0, Cache(self._cachelen))

    def save(self, input, output=None, packed=True):
        if output is None:
            output = {}
        self._save(input, output, packed, set())
        return output

# mix-in for all generators of nullable types
class Masked(object):
    maskdtype = numpy.dtype(numpy.int32)
    maskedvalue = -1

    def __init__(self, mask, maskidx, packmask):
        self.mask = mask
        self.maskidx = maskidx
        self.packmask = packmask

    def _fallback(self, arrays, cache, name, err):
        # packmask = numpy.packbits(mask != self.maskedvalue)
        if name == self.mask:
            try:
                packmask = arrays[self.packmask]
            except KeyError:
                raise err
            else:
                if not isinstance(packmask, numpy.ndarray):
                    packmask = numpy.array(packmask, dtype=numpy.dtype(numpy.uint8))
                if packmask.dtype != numpy.dtype(numpy.uint8):
                    raise TypeError("arrays[{0}].dtype is {1} but expected {2}".format(repr(self.packmask), packmask.dtype, numpy.dtype(numpy.uint8)))
                if packmask.shape[1:] != ():
                    raise TypeError("arrays[{0}].shape[1:] is {1} but expected ()".format(repr(self.packmask), packmask.shape[1:]))
                unmasked = numpy.unpackbits(packmask).view(numpy.bool_)
                mask = numpy.empty(len(unmasked), dtype=self.maskdtype)
                mask[unmasked] = numpy.arange(unmasked.sum(), dtype=self.maskdtype)
                mask[~unmasked] = self.maskedvalue
                return mask                             # may have excess self.maskedvalue values at the end due to zero-padding in numpy.packbits
        else:
            return self.__class__.__bases__[1]._fallback(self, arrays, cache, name, err)

    def _generate(self, arrays, index, cache):
        value = self._getarray(arrays, self.mask, cache, self.maskidx, self.maskdtype, ())[index]
        if value == self.maskedvalue:
            return None
        else:
            # otherwise, the value is the index for packed data
            return self.__class__.__bases__[1]._generate(self, arrays, value, cache)

    def _save(self, input, output, packed, memo):
        mask = self._toarray(input[self.mask], self.maskdtype)
        if packed:
            output[self.packmask] = numpy.packbits(mask != self.maskedvalue)
        else:
            output[self.mask] = mask
        self.__class__.__bases__[1]._save(self, input, output, packed, memo)

################################################################ Primitives

class PrimitiveGenerator(Generator):
    def __init__(self, data, dataidx, dtype, dims, name, derivedname, schema):
        self.data = data
        self.dataidx = dataidx
        self.dtype = dtype
        self.dims = dims
        Generator.__init__(self, name, derivedname, schema)

    def _generate(self, arrays, index, cache):
        return self._getarray(arrays, self.data, cache, self.dataidx, self.dtype, self.dims)[index]

    def _save(self, input, output, packed, memo):
        output[self.data] = self._toarray(input[self.data], self.dtype)

class MaskedPrimitiveGenerator(Masked, PrimitiveGenerator):
    def __init__(self, mask, maskidx, packmask, data, dataidx, dtype, dims, name, derivedname, schema):
        Masked.__init__(self, mask, maskidx, packmask)
        PrimitiveGenerator.__init__(self, data, dataidx, dtype, dims, name, derivedname, schema)

################################################################ Lists

class ListGenerator(Generator):
    posdtype = numpy.dtype(numpy.int32)

    def __init__(self, starts, startsidx, stops, stopsidx, counts, content, name, derivedname, schema):
        self.starts = starts
        self.startsidx = startsidx
        self.stops = stops
        self.stopsidx = stopsidx
        self.counts = counts
        self.content = content
        Generator.__init__(self, name, derivedname, schema)

    def _fallback(self, arrays, cache, name, err):
        # counts = stops - starts
        if name == self.starts:
            try:
                counts = arrays[self.counts]
            except KeyError:
                raise err
            else:
                if getattr(counts, "dtype", None) != self.posdtype:
                    counts = numpy.array(counts, dtype=self.posdtype)
                if counts.shape[1:] != ():
                    raise TypeError("arrays[{0}].shape[1:] is {1} but expected ()".format(repr(self.counts), counts.shape[1:]))
                offsets = numpy.empty(len(counts) + 1, dtype=self.posdtype)
                offsets[0] = 0
                offsets[1:] = numpy.cumsum(counts)
                return offsets                          # offsets is a starts array with an excess value (total length) at the end
        elif name == self.stops:
            # already filled starts with offsets
            return cache.arraylist[self.startsidx][1:]  # take off the initial zero (now we need that excess value at the end!)
        else:
            raise err

    def _generate(self, arrays, index, cache):
        starts = self._getarray(arrays, self.starts, cache, self.startsidx, self.posdtype, ())
        stops  = self._getarray(arrays, self.stops,  cache, self.stopsidx,  self.posdtype, ())
        return oamap.proxy.ListProxy(self, arrays, cache, starts[index], 1, stops[index] - starts[index])

    def _save(self, input, output, packed, memo):
        starts = self._toarray(input[self.starts], self.posdtype)
        stops  = self._toarray(input[self.stops], self.posdtype)
        if packed and starts[0] == 0 and numpy.array_equal(starts[1:], stops[:-1]):
            output[self.counts] = stops - starts
        else:
            output[self.starts] = starts
            output[self.stops]  = stops
        self.content._save(input, output, packed, memo)

class MaskedListGenerator(Masked, ListGenerator):
    def __init__(self, mask, maskidx, packmask, starts, startsidx, stops, stopsidx, counts, content, name, derivedname, schema):
        Masked.__init__(self, mask, maskidx, packmask)
        ListGenerator.__init__(self, starts, startsidx, stops, stopsidx, counts, content, name, derivedname, schema)

################################################################ Unions

class UnionGenerator(Generator):
    tagdtype = numpy.dtype(numpy.int8)
    offsetdtype = numpy.dtype(numpy.int32)

    def __init__(self, tags, tagsidx, offsets, offsetsidx, possibilities, name, derivedname, schema):
        self.tags = tags
        self.tagsidx = tagsidx
        self.offsets = offsets
        self.offsetsidx = offsetsidx
        self.possibilities = possibilities
        Generator.__init__(self, name, derivedname, schema)

    def _fallback(self, arrays, cache, name, err):
        if name == self.offsets:
            # already filled tags
            tags = cache.arraylist[self.tagsidx]
            offsets = numpy.empty(len(tags), dtype=self.offsetdtype)
            for tag in range(len(self.possibilities)):
                hastag = (tags == tag)
                offsets[hastag] = numpy.arange(hastag.sum(), dtype=self.offsetdtype)
            # assume that every value in tags is associated with one of the possibilities (i.e. tags is well-formed)
            return offsets
        else:
            raise err

    def _generate(self, arrays, index, cache):
        tags    = self._getarray(arrays, self.tags,    cache, self.tagsidx,    self.tagdtype,    ())
        offsets = self._getarray(arrays, self.offsets, cache, self.offsetsidx, self.offsetdtype, ())
        return self.possibilities[tags[index]]._generate(arrays, offsets[index], cache)

    def _save(self, input, output, packed, memo):
        output[self.tags] = self._toarray(input[self.tags], self.tagdtype)
        if not packed:
            output[self.offsets] = self._toarray(input[self.offsets], self.offsetdtype)
        for x in self.possibilities:
            x._save(input, output, packed, memo)

class MaskedUnionGenerator(Masked, UnionGenerator):
    def __init__(self, mask, maskidx, packmask, tags, tagsidx, offsets, offsetsidx, possibilities, name, derivedname, schema):
        Masked.__init__(self, mask, maskidx, packmask)
        UnionGenerator.__init__(self, tags, tagsidx, offsets, offsetsidx, possibilities, name, derivedname, schema)

################################################################ Records

class RecordGenerator(Generator):
    def __init__(self, fields, name, derivedname, schema):
        self.fields = fields
        Generator.__init__(self, name, derivedname, schema)

    def _generate(self, arrays, index, cache):
        return oamap.proxy.RecordProxy(self, arrays, cache, index)

    def _save(self, input, output, packed, memo):
        for x in self.fields.values():
            x._save(input, output, packed, memo)

class MaskedRecordGenerator(Masked, RecordGenerator):
    def __init__(self, mask, maskidx, packmask, fields, name, derivedname, schema):
        Masked.__init__(self, mask, maskidx, packmask)
        RecordGenerator.__init__(self, fields, name, derivedname, schema)

################################################################ Tuples

class TupleGenerator(Generator):
    def __init__(self, types, name, derivedname, schema):
        self.types = types
        Generator.__init__(self, name, derivedname, schema)

    def _generate(self, arrays, index, cache):
        return oamap.proxy.TupleProxy(self, arrays, cache, index)

    def _save(self, input, output, packed, memo):
        for x in self.types:
            x._save(input, output, packed, memo)

class MaskedTupleGenerator(Masked, TupleGenerator):
    def __init__(self, mask, maskidx, packmask, types, name, derivedname, schema):
        Masked.__init__(self, mask, maskidx, packmask)
        TupleGenerator.__init__(self, types, name, derivedname, schema)

################################################################ Pointers

class PointerGenerator(Generator):
    posdtype = numpy.dtype(numpy.int32)

    def __init__(self, positions, positionsidx, target, name, derivedname, schema):
        self.positions = positions
        self.positionsidx = positionsidx
        self.target = target
        Generator.__init__(self, name, derivedname, schema)

    def _generate(self, arrays, index, cache):
        positions = self._getarray(arrays, self.positions, cache, self.positionsidx, self.posdtype, ())
        return self.target._generate(arrays, positions[index], cache)

    def _save(self, input, output, packed, memo):
        if id(self) not in memo:
            memo.add(id(self))
            output[self.positions] = self._toarray(input[self.positions], self.posdtype)
            self.target._save(input, output, packed, memo)

class MaskedPointerGenerator(Masked, PointerGenerator):
    def __init__(self, mask, maskidx, packmask, positions, positionsidx, target, name, derivedname, schema):
        Masked.__init__(self, mask, maskidx, packmask)
        PointerGenerator.__init__(self, positions, positionsidx, target, name, derivedname, schema)

################################################################ for extensions: domain-specific and user

class ExtendedGenerator(Generator):
    # extensions *must* override pattern, _generate, and degenerate, *may* override matches
    pattern = None

    # default implementation: generate a given type of proxy
    proxyclass = None
    def _generate(self, arrays, index, cache):
        return self.proxyclass(self, arrays, cache, index)

    # default implementation: do nothing; generic type is the same as extended type
    def degenerate(self, obj):
        return obj

    def __init__(self, genericclass, *args):
        self.generic = genericclass(*args)

    @property
    def name(self):
        return self.generic.name

    @property
    def derivedname(self):
        return self.generic.derivedname

    @property
    def schema(self):
        return self.generic.schema

    @classmethod
    def matches(cls, schema):
        def recurse(pattern, schema):
            if isinstance(pattern, basestring):
                return schema["type"] == "primitive" and schema["dtype"] == pattern
            else:
                assert isinstance(pattern, dict) and "type" in pattern

                if "nullable" in pattern and pattern["nullable"] != schema["nullable"]:
                    return False
                if "name" in pattern and pattern["name"] != schema["name"]:
                    return False
                if "label" in pattern and pattern["label"] != schema["label"]:
                    return False

                if pattern["type"] == "primitive":
                    if "dtype" in pattern and pattern["dtype"] != schema["dtype"]:
                        return False
                    if "dims" in pattern and pattern["dims"] != schema["dims"]:
                        return False
                    return True

                elif pattern["type"] == "list":
                    if "content" in pattern:
                        return recurse(pattern["content"], schema["content"])
                    return True

                elif pattern["type"] == "union":
                    if "possibilities" in pattern:
                        return len(pattern["possibilities"]) == len(schema["possibilities"]) and all(recurse(x, y) for x, y in zip(pattern["possibilities"], schema["possibilities"]))
                    return True

                elif pattern["type"] == "record":
                    if "fields" in pattern:
                        if isinstance(pattern["fields"], dict):
                            return set(pattern["fields"].keys()).issubset(set(n for n, x in schema["fields"])) and all(recurse(px, [sx for sn, sx in schema["fields"] if pn == sn][0]) for pn, px in pattern["fields"].items())
                        else:
                            return len(pattern["fields"]) == len(schema["fields"]) and all(pn == sn and recurse(px, sx) for (pn, px), (sn, sx) in zip(pattern["fields"], schema["fields"]))
                    return True

                elif pattern["type"] == "tuple":
                    if "types" in pattern:
                        return len(pattern["types"]) == len(schema["types"]) and all(recurse(x, y) for x, y in zip(pattern["types"], schema["types"]))
                    return True

                elif pattern["type"] == "pointer":
                    if "target" in pattern:
                        return recurse(pattern["target"], schema["target"])
                    return True

                else:
                    assert pattern["type"] in ("primitive", "list", "union", "record", "tuple", "pointer")

        return recurse(cls.pattern, schema.tojson(explicit=True))

    def _fallback(self, arrays, cache, name, err):
        return self.generic._fallback(arrays, cache, name, err)

    def _save(self, input, output, packed, memo):
        self.generic._save(input, output, packed, memo)
        
################################################################ for assigning unique strings to types (used to distinguish Numba types)

def _firstindex(generator):
    if isinstance(generator, Masked):
        return generator.maskidx
    elif isinstance(generator, PrimitiveGenerator):
        return generator.dataidx
    elif isinstance(generator, ListGenerator):
        return generator.startsidx
    elif isinstance(generator, UnionGenerator):
        return generator.tagsidx
    elif isinstance(generator, RecordGenerator):
        for g in generator.fields.values():
            return _firstindex(g)
        return -1
    elif isinstance(generator, TupleGenerator):
        for g in generator.types:
            return _firstindex(g)
        return -1
    elif isinstance(generator, PointerGenerator):
        return generator.positionsidx
    else:
        raise AssertionError("unrecognized generator type: {0}".format(generator))

def _uniquestr(generator, memo):
    if id(generator) not in memo:
        memo.add(id(generator))
        givenname = "nil" if generator.name is None else repr(generator.name)

        if isinstance(generator, PrimitiveGenerator):
            generator._uniquestr = "(P {0} {1} ({2}) {3} {4})".format(givenname, repr(str(generator.dtype)), " ".join(map(repr, generator.dims)), generator.dataidx, repr(generator.data))

        elif isinstance(generator, MaskedPrimitiveGenerator):
            generator._uniquestr = "(P {0} {1} ({2}) {3} {4} {5} {6} {7})".format(givenname, repr(str(generator.dtype)), " ".join(map(repr, generator.dims)), generator.maskidx, repr(generator.mask), repr(generator.packmask), generator.dataidx, repr(generator.data))

        elif isinstance(generator, ListGenerator):
            _uniquestr(generator.content, memo)
            generator._uniquestr = "(L {0} {1} {2} {3} {4} {5} {6})".format(givenname, generator.startsidx, repr(generator.starts), generator.stopsidx, repr(generator.stops), repr(generator.counts), generator.content._uniquestr)

        elif isinstance(generator, MaskedListGenerator):
            _uniquestr(generator.content, memo)
            generator._uniquestr = "(L {0} {1} {2} {3} {4} {5} {6} {7})".format(givenname, generator.maskidx, repr(generator.mask), repr(generator.packmask), generator.startsidx, repr(generator.starts), generator.stopsidx, repr(generator.stops), repr(generator.counts), generator.content._uniquestr)

        elif isinstance(generator, UnionGenerator):
            for t in generator.possibilities:
                _uniquestr(t, memo)
            generator._uniquestr = "(U {0} {1} {2} {3} {4} ({5}))".format(givenname, generator.tagsidx, repr(generator.tags), generator.offsetsidx, repr(generator.offsets), " ".join(x._uniquestr for x in generator.possibilities))

        elif isinstance(generator, MaskedUnionGenerator):
            for t in generator.possibilities:
                _uniquestr(t, memo)
            generator._uniquestr = "(U {0} {1} {2} {3} {4} {5} {6} {7} ({8}))".format(givenname, generator.maskidx, repr(generator.mask), repr(generator.packmask), generator.tagsidx, repr(generator.tags), generator.offsetsidx, repr(generator.offsets), " ".join(x._uniquestr for x in generator.possibilities))

        elif isinstance(generator, RecordGenerator):
            for t in generator.fields.values():
                _uniquestr(t, memo)
            generator._uniquestr = "(R {0} ({1}))".format(givenname, " ".join("({0} . {1})".format(repr(n), t._uniquestr) for n, t in generator.fields.items()))

        elif isinstance(generator, MaskedRecordGenerator):
            for t in generator.fields.values():
                _uniquestr(t, memo)
            generator._uniquestr = "(R {0} {1} {2} {3} ({4}))".format(givenname, generator.maskidx, repr(generator.mask), repr(generator.packmask), " ".join("({0} . {1})".format(repr(n), t._uniquestr) for n, t in generator.fields.items()))

        elif isinstance(generator, TupleGenerator):
            for t in generator.types:
                _uniquestr(t, memo)
            generator._uniquestr = "(T {0} ({1}))".format(givenname, " ".join(t._uniquestr for t in generator.types))

        elif isinstance(generator, MaskedTupleGenerator):
            for t in generator.types:
                _uniquestr(t, memo)
            generator._uniquestr = "(T {0} {1} {2} {3} ({4}))".format(givenname, generator.maskidx, repr(generator.mask), repr(generator.packmask), " ".join(t._uniquestr for t in generator.types))

        elif isinstance(generator, PointerGenerator):
            _uniquestr(generator.target, memo)
            if generator._internal:
                target = _firstindex(generator.target)
            else:
                target = generator.target._uniquestr
            generator._uniquestr = "(X {0} {1} {2} {3})".format(givenname, generator.positionsidx, repr(generator.positions), target)

        elif isinstance(generator, MaskedPointerGenerator):
            _uniquestr(generator.target, memo)
            if generator._internal:
                target = _firstindex(generator.target)
            else:
                target = generator.target._uniquestr
            generator._uniquestr = "(X {0} {1} {2} {3} {4} {5} {6})".format(givenname, generator.maskidx, repr(generator.mask), repr(generator.packmask), generator.positionsidx, repr(generator.positions), target)

        elif isinstance(generator, ExtendedGenerator):
            _uniquestr(generator.generic, memo)
            generator._uniquestr = "({0} {1})".format(generator.__class__.__name__, generator.generic._uniquestr)

        else:
            raise AssertionError("unrecognized generator type: {0}".format(generator))



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import codecs
import sys

import numpy

import oamap.generator

class _GenerateBytes(object):
    py3 = sys.version_info[0] >= 3

    def _generatebytes(self, arrays, index, cache):
        if self.schema.nullable:
            value = self._getarray(arrays, self.generic.mask, cache, self.generic.maskidx, self.generic.maskdtype, ())[index]
            if value == self.generic.maskedvalue:
                return None
            else:
                index = value

        listgen = self.generic
        primgen = self.generic.content

        starts = self._getarray(arrays, listgen.starts, cache, listgen.startsidx, listgen.posdtype, ())
        stops  = self._getarray(arrays, listgen.stops,  cache, listgen.stopsidx,  listgen.posdtype, ())
        data   = self._getarray(arrays, primgen.data,   cache, primgen.dataidx,   primgen.dtype, ())
        array  = data[starts[index]:stops[index]]

        if isinstance(array, numpy.ndarray):
            return array.tostring()
        elif self.py3:
            return bytes(array)
        else:
            return "".join(map(chr, array))

    def degenerate(self, obj):
        if obj is None:
            return obj

        elif self.py3:
            if isinstance(obj, bytes):
                return obj
            else:
                return codecs.utf_8_encode(obj)[0]

        else:
            if isinstance(obj, str):
                return map(ord, obj)
            else:
                return map(ord, codecs.utf_8_encode(obj)[0])

class ByteStringGenerator(_GenerateBytes, oamap.generator.ExtendedGenerator):
    pattern = {"name": "ByteString", "type": "list", "content": {"type": "primitive", "dtype": "uint8", "dims": [], "nullable": False}}

    def _generate(self, arrays, index, cache):
        return self._generatebytes(arrays, index, cache)
            
class UTF8StringGenerator(_GenerateBytes, oamap.generator.ExtendedGenerator):
    pattern = {"name": "UTF8String", "type": "list", "content": {"type": "primitive", "dtype": "uint8", "dims": [], "nullable": False}}

    def _generate(self, arrays, index, cache):
        out = self._generatebytes(arrays, index, cache)
        if out is None:
            return out
        else:
            return codecs.utf_8_decode(out)[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path

from setuptools import find_packages
from setuptools import setup

def get_version():
    g = {}
    exec(open(os.path.join("oamap", "version.py")).read(), g)
    return g["__version__"]

setup(name = "oamap",
      version = get_version(),
      packages = find_packages(exclude = ["tests"]),
      scripts = [],
      description = "Tools for analyzing hierarchically nested, columnar data without deserialization.",
      long_description = """Data analysts are often faced with a choice between speed and flexibility. Tabular data, such as SQL tables, can be processed rapidly enough for a truly interactive analysis session, but hierarchically nested data, such as JSON, is better at representing relationships within complex data models. In some domains (such as particle physics), we want to perform calculations on JSON-like structures at the speed of SQL.

The key to high throughput for large datasets, particularly ones with many more attributes than are typically accessed in a single pass, is laying out the data in "columns." All values of an attribute should be contiguous in disk or memory because data are paged from one cache to the next in locally contiguous blocks. The `ROOT <https://root.cern/>`_ and `Parquet <http://parquet.apache.org/>`_ file formats represent hierarchically nested data in a columnar form on disk, and `Apache Arrow <https://arrow.apache.org/>`_ is an emerging standard for sharing hierarchically nested data in memory. However, data from these formats are usually deserialized into conventional data structures before processing, which limits performance (see `this talk <https://youtu.be/jvt4v2LTGK0>`_ and `this paper <https://arxiv.org/abs/1711.01229>`_).

OAMap is a toolset for computing arbitrary functions on hierarchical, columnar data without deserialization. The name stands for Object Array Mapping, in analogy with the `Object Relational Mapping (ORM) <https://en.wikipedia.org/wiki/Object-relational_mapping>`_ interface to some databases. Users (data analysts) write functions on JSON-like objects that OAMap compiles into operations on the underlying arrays, similar to the way that ORM converts object-oriented code into SQL. The difference is that mapping objects to non-relational arrays permits bare-metal performance (giving up some traditional database features).

OAMap is a Python library on top of which high-level analysis software may be built. It focuses on mapping an object-oriented view of data onto columnar arrays. `Numpy <http://www.numpy.org/>`_ is OAMap's only strict dependency, though OAMap objects are also implemented as `Numba extensions <http://numba.pydata.org/numba-doc/dev/extending/index.html>`_, so they may be used in `Numba's <http://numba.pydata.org/>`_ JIT-compiled functions at speeds that match or exceed hand-written C code. OAMap is unopinionated about the source of its columnar arrays, allowing for a variety of backends. See `the walkthrough <https://github.com/diana-hep/oamap/blob/master/README.rst#walkthrough>`_ for more.

Also, a similar object array mapping could be implemented in any language--- Python was chosen only for its popularity among data analysts.""",
      author = "Jim Pivarski (DIANA-HEP)",
      author_email = "pivarski@fnal.gov",
      maintainer = "Jim Pivarski (DIANA-HEP)",
      maintainer_email = "pivarski@fnal.gov",
      url = "https://github.com/diana-hep/oamap",
      download_url = "https://github.com/diana-hep/oamap/releases",
      license = "BSD 3-clause",
      test_suite = "tests",
      install_requires = ["numpy"],
      tests_require = [],
      classifiers = [
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      platforms = "Any",
      )



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
import numbers

import numpy

import oamap.schema
from oamap.schema import OrderedDict

################################################################ inferring schemas from data

def jsonconventions(value):
    def recurse(value, memo):
        if id(value) in memo:
            return memo[id(value)]
        memo[id(value)] = value

        if isinstance(value, dict) and len(value) == 2 and set(value.keys()) == set(["real", "imag"]) and all(isinstance(x, (int, float)) for x in value.values()):
            return value["real"] + value["imag"]*1j
        elif value == "inf":
            return float("inf")
        elif value == "-inf":
            return float("-inf")
        elif value == "nan":
            return float("nan")
        elif isinstance(value, list):
            return [recurse(x, memo) for x in value]
        elif isinstance(value, dict):
            return dict((n, recurse(x, memo)) for n, x in value.items())
        else:
            return value

    return recurse(value, {})

def fromjson(obj, limititems=None):
    return fromdata(jsonconventions(obj), limititems=limititems)

def fromdata(obj, limititems=None):
    if limititems is None or (isinstance(limititems, numbers.Integral) and limititems >= 0):
        pass
    else:
        raise TypeError("limititems must be None or a non-negative integer, not {0}".format(limititems))

    class Intermediate(object):
        def __init__(self, nullable):
            self.nullable = nullable

    class Unknown(Intermediate):
        def resolve(self):
            raise TypeError("could not resolve a type (e.g. all examples of a List-typed attribute are empty, can't determine its content type)")

    class Boolean(Intermediate):
        def resolve(self):
            return oamap.schema.Primitive(numpy.dtype(numpy.bool_), nullable=self.nullable)

    class Number(Intermediate):
        max_uint8 = numpy.iinfo(numpy.uint8).max
        max_uint16 = numpy.iinfo(numpy.uint16).max
        max_uint32 = numpy.iinfo(numpy.uint32).max
        max_uint64 = numpy.iinfo(numpy.uint64).max
        min_int8 = numpy.iinfo(numpy.int8).min
        max_int8 = numpy.iinfo(numpy.int8).max
        min_int16 = numpy.iinfo(numpy.int16).min
        max_int16 = numpy.iinfo(numpy.int16).max
        min_int32 = numpy.iinfo(numpy.int32).min
        max_int32 = numpy.iinfo(numpy.int32).max
        min_int64 = numpy.iinfo(numpy.int64).min
        max_int64 = numpy.iinfo(numpy.int64).max
        def __init__(self, nullable, min, max, whole, real):
            Intermediate.__init__(self, nullable)
            self.min = min
            self.max = max
            self.whole = whole
            self.real = real
        def resolve(self):
            if self.whole:
                if self.min >= 0:
                    if self.max <= self.max_uint8:
                        t = numpy.uint8
                    elif self.max <= self.max_uint16:
                        t = numpy.uint16
                    elif self.max <= self.max_uint32:
                        t = numpy.uint32
                    elif self.max <= self.max_uint64:
                        t = numpy.uint64
                    else:
                        t = numpy.float64
                else:
                    if self.min_int8 <= self.min and self.max <= self.max_int8:
                        t = numpy.int8
                    elif self.min_int16 <= self.min and self.max <= self.max_int16:
                        t = numpy.int16
                    elif self.min_int32 <= self.min and self.max <= self.max_int32:
                        t = numpy.int32
                    elif self.min_int64 <= self.min and self.max <= self.max_int64:
                        t = numpy.int64
                    else:
                        t = numpy.float64
            elif self.real:
                t = numpy.float64
            else:
                t = numpy.complex128
            return oamap.schema.Primitive(numpy.dtype(t), nullable=self.nullable)

    class IntermediateList(Intermediate):
        def __init__(self, nullable, content):
            Intermediate.__init__(self, nullable)
            self.content = content
        def resolve(self):
            return oamap.schema.List(self.content.resolve(), nullable=self.nullable)

    class IntermediateRecord(Intermediate):
        def __init__(self, nullable, fields, name):
            Intermediate.__init__(self, nullable)
            self.fields = fields
            self.name = name
        def resolve(self):
            return oamap.schema.Record(dict((n, x.resolve()) for n, x in self.fields.items()), nullable=self.nullable, name=self.name)

    class IntermediateTuple(Intermediate):
        def __init__(self, nullable, types):
            Intermediate.__init__(self, nullable)
            self.types = types
        def resolve(self):
            return oamap.schema.Tuple([x.resolve() for x in self.types], nullable=self.nullable)

    # Unions are special for type-inference
    class IntermediateUnion(Intermediate):
        def __init__(self, nullable, possibilities):
            Intermediate.__init__(self, nullable)
            self.possibilities = possibilities
        def resolve(self):
            return oamap.schema.Union([x.resolve() for x in self.possibilities], nullable=self.nullable)

    # no Pointers in type-inference (we'd have to keep a big map of *everything*!)

    def flatten(possibilities):
        return [y for x in possibilities if isinstance(x, IntermediateUnion) for y in x.possibilities] + [x for x in possibilities if not isinstance(x, IntermediateUnion)]

    def unify2(x, y):
        nullable = x.nullable or y.nullable

        if isinstance(x, Unknown) and isinstance(y, Unknown):
            return Unknown(nullable)

        elif isinstance(x, Unknown):
            y.nullable = nullable
            return y

        elif isinstance(y, Unknown):
            x.nullable = nullable
            return x

        elif isinstance(x, Boolean) and isinstance(y, Boolean):
            return Boolean(nullable)

        elif isinstance(x, Number) and isinstance(y, Number):
            return Number(nullable, min(x.min, y.min), max(x.max, y.max), x.whole and y.whole, x.real and y.real)

        elif isinstance(x, IntermediateList) and isinstance(y, IntermediateList):
            return IntermediateList(nullable, unify2(x.content, y.content))

        elif isinstance(x, IntermediateRecord) and isinstance(y, IntermediateRecord) and set(x.fields) == set(y.fields) and (x.name is None or y.name is None or x.name == y.name):
            return IntermediateRecord(nullable, dict((n, unify2(x.fields[n], y.fields[n])) for n in x.fields), name=(y.name if x.name is None else x.name))

        elif isinstance(x, IntermediateTuple) and isinstance(y, IntermediateTuple) and len(x.types) == len(y.types):
            return IntermediateTuple(nullable, [unify2(xi, yi) for xi, yi in zip(x.types, y.types)])

        elif isinstance(x, IntermediateUnion) and isinstance(y, IntermediateUnion):
            return unify(x.possibilities + y.possibilities)

        elif isinstance(x, IntermediateUnion):
            return unify(x.possibilities + [y])

        elif isinstance(y, IntermediateUnion):
            return unify([x] + y.possibilities)

        else:
            # can't be unified
            return IntermediateUnion(nullable, flatten([x, y]))

    def unify(possibilities):
        if len(possibilities) == 0:
            return Unknown(False)

        elif len(possibilities) == 1:
            return possibilities[0]

        elif len(possibilities) == 2:
            return unify2(possibilities[0], possibilities[1])

        else:
            distinct = []
            for x in flatten(possibilities):
                found = False

                for i, y in enumerate(distinct):
                    merged = unify2(x, y)
                    if not isinstance(merged, IntermediateUnion):
                        distinct[i] = merged
                        found = True
                        break

                if not found:
                    distinct.append(x)

            if len(distinct) == 1:
                return distinct[0]
            else:
                return IntermediateUnion(False, flatten(distinct))

    def buildintermediate(obj, limititems, memo):
        if id(obj) in memo:
            raise ValueError("cyclic reference in Python object at {0} (Pointer types cannot be inferred)".format(obj))

        # by copying, rather than modifying in-place (memo.add), we find cyclic references, rather than DAGs
        memo = memo.union(set([id(obj)]))

        if obj is None:
            return Unknown(True)

        elif obj is False or obj is True:
            return Boolean(False)

        elif isinstance(obj, numbers.Integral):
            return Number(False, int(obj), int(obj), True, True)

        elif isinstance(obj, numbers.Real):
            return Number(False, float(obj), float(obj), False, True)

        elif isinstance(obj, numbers.Complex):
            return Number(False, float("-inf"), float("inf"), False, False)

        elif isinstance(obj, dict):
            return IntermediateRecord(False, dict((n, buildintermediate(x, limititems, memo)) for n, x in obj.items()), None)

        elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
            # this is a namedtuple; interpret it as a Record, rather than a Tuple
            return IntermediateRecord(False, dict((n, buildintermediate(getattr(obj, n), limititems, memo)) for n in obj._fields), obj.__class__.__name__)

        elif isinstance(obj, tuple):
            return IntermediateTuple(False, [buildintermediate(x, limititems, memo) for x in obj])

        else:
            try:
                limited = []
                for x in obj:
                    if limititems is None or len(limited) < limititems:
                        limited.append(x)
                    else:
                        break
            except TypeError:
                # not iterable, so interpret it as a Record
                return IntermediateRecord(False, dict((n, buildintermediate(getattr(obj, n), limititems, memo)) for n in dir(obj) if not n.startswith("_") and not callable(getattr(obj, n))), obj.__class__.__name__)
            else:
                # iterable, so interpret it as a List
                return IntermediateList(False, unify([buildintermediate(x, None, memo) for x in obj]))

    return buildintermediate(obj, limititems, set()).resolve()

################################################################ inferring schemas from a namespace

def fromnames(arraynames, prefix="object", delimiter="-"):
    def filter(arraynames, prefix):
        return [x for x in arraynames if x.startswith(prefix)]
    
    def recurse(arraynames, prefix, byname, internalpointers):
        prefixdelimiter = prefix + delimiter
        name = None
        for n in arraynames:
            if n.startswith(prefixdelimiter):
                if n[len(prefixdelimiter)] == "N":
                    match = oamap.schema.Schema._identifier.match(n[len(prefixdelimiter) + 1:])
                    if match is not None:
                        name = match.group(0)
                        break

        if name is not None:
            prefix = prefixdelimiter + "N" + name
            prefixdelimiter = prefix + delimiter
            
        mask      = prefixdelimiter + "M"
        packmask  = prefixdelimiter + "m"
        starts    = prefixdelimiter + "B"
        stops     = prefixdelimiter + "E"
        counts    = prefixdelimiter + "c"
        content   = prefixdelimiter + "L"
        tags      = prefixdelimiter + "T"
        offsets   = prefixdelimiter + "O"
        uniondata = prefixdelimiter + "U"
        field     = prefixdelimiter + "F"
        positions = prefixdelimiter + "P"
        external  = prefixdelimiter + "X"
        primitive = prefixdelimiter + "D"

        nullable = mask in arraynames
        if not nullable:
            mask = None
            packmask = None

        if (starts in arraynames and stops in arraynames) or counts in arraynames:
            byname[prefix] = None
            byname[prefix] = oamap.schema.List(recurse(filter(arraynames, content), content, byname, internalpointers), nullable=nullable, starts=None, stops=None, counts=None, mask=None, packmask=None, name=name, doc=None)

        elif tags in arraynames:
            possibilities = []
            while True:
                possibility = uniondata + repr(len(possibilities))
                if any(x.startswith(possibility) for x in arraynames):
                    possibilities.append(possibility)
                else:
                    break
            byname[prefix] = None
            byname[prefix] = oamap.schema.Union([recurse(filter(arraynames, x), x, byname, internalpointers) for x in possibilities], nullable=nullable, tags=None, offsets=None, mask=None, packmask=None, name=name, doc=None)

        elif any(x.startswith(field) for x in arraynames):
            pattern = re.compile("^" + field + "(" + oamap.schema.Schema._identifier.pattern + ")")
            fields = {}
            for x in arraynames:
                matches = pattern.match(x)
                if matches is not None:
                    if matches.group(1) not in fields:
                        fields[matches.group(1)] = []
                    fields[matches.group(1)].append(x)

            types = []
            while True:
                tpe = field + repr(len(types))
                if any(x.startswith(tpe) for x in arraynames):
                    types.append(tpe)
                else:
                    break

            if len(fields) >= 0 and len(types) == 0:
                byname[prefix] = oamap.schema.Record(oamap.schema.OrderedDict([(n, recurse(fields[n], field + n, byname, internalpointers)) for n in sorted(fields)]), nullable=nullable, mask=None, packmask=None, name=name, doc=None)
            elif len(fields) == 0 and len(types) > 0:
                byname[prefix] = oamap.schema.Tuple([recurse(filter(arraynames, n), n, byname, internalpointers) for n in types], nullable=nullable, mask=None, packmask=None, name=name, doc=None)
            else:
                raise KeyError("ambiguous set of array names: may be Record or Tuple at {0}".format(repr(prefix)))

        elif any(x.startswith(positions) for x in arraynames):
            if positions in arraynames:
                # external
                byname2 = {}
                internalpointers2 = []
                target = finalize(recurse(filter(arraynames, external), external, byname2, internalpointers2), byname2, internalpointers2)
                byname[prefix] = oamap.schema.Pointer(target, nullable=nullable, positions=None, mask=None, packmask=None, name=name, doc=None)

            else:
                # internal
                matches = [x[len(positions) + 1:] for x in arraynames if x.startswith(positions)]
                if len(matches) != 1:
                    raise KeyError("ambiguous set of array names: more than one internal Pointer at {0}".format(repr(prefix)))
                target = None   # placeholder! see finalize
                byname[prefix] = oamap.schema.Pointer(target, nullable=nullable, positions=None, mask=None, packmask=None, name=name, doc=None)
                internalpointers.append((byname[prefix], matches[0]))

        elif any(x.startswith(primitive) for x in arraynames):
            matches = [x[len(primitive):] for x in arraynames if x.startswith(primitive)]
            if len(matches) != 1:
                raise KeyError("ambiguous set of array names: more than one Primitive at {0}".format(repr(prefix)))
            info = matches[0].split(delimiter)
            if info[0].startswith("S"):
                dtype = numpy.dtype(info[0])
            elif info[0].isupper():
                dtype = numpy.dtype(">" + info[0].lower())
            else:
                dtype = numpy.dtype("<" + info[0])
            dims = tuple(int(x) for x in info[1:])
            byname[prefix] = oamap.schema.Primitive(dtype, dims=dims, nullable=nullable, data=None, mask=None, packmask=None, name=name, doc=None)

        else:
            raise KeyError("missing array names: nothing found as {0} contents".format(repr(prefix)))

        return byname[prefix]

    def finalize(out, byname, internalpointers):
        for pointer, targetname in internalpointers:
            if targetname in byname:
                pointer.target = byname[targetname]
            else:
                raise KeyError("Pointer's internal target is {0}, but there is no object with that prefix".format(repr(targetname)))
        return out

    byname = {}
    internalpointers = []
    return finalize(recurse(filter(arraynames, prefix), prefix, byname, internalpointers), byname, internalpointers)
